/*
Batterieberwachung
Grundlage: TinySA,MPPT,
PB0 - Extra-Last
PB1 - Funky
PB2 - Alarm
PB3 - UBat
PB4 - Strom

Funktionen: 
- Anzeige von U und I, P, Entladungstiefe, Temperatur
- Steuerung optionaler Verbraucher (Auto,Klima)
- Alarmverarbeitung

Debugverlauf:
mlibc nicht eingebunden
falscher Sockel fr ina118p
Interupt-Quelle beim Senden und Empfangen deaktivieren 
Totzeit nach Senden einrichten, C an ina118
TK vom Stromsensor: 0.6A bei 32Grad
Fehler bei neg. Strmen bei drucke()
print-Funktion verschlankt
*/
#include <avr/io.h> 
#include <avr/wdt.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <string.h>
#include "board.h"			

#define Vbg_mV 			1060L		// 1082 059interne Referenz
#define UBAT_OS_mV		100L		// Feinjustage
#define I_OS_mA 		10700L		// Fehler der Messbrcke
#define ISENS_uV_pro_mA 25L			// V pro mA
#define I_Adjust		945L		// Feinkorrektur pro 1000
#define IBAT_CHR_AL 	-75000L		// max Ladestrom
#define IBAT_DCH_AL 	100000L		// max Enladestrom
#define UBAT_MAX_AL 	55000L		// obere Grenze
#define UBAT_MIN_AL 	45000L		// untere Grenze
#define TEMP_AL 		40000L		// 40C Interner-Temperaturalarm
#define UBAT_FULL 		52800L		// knapp unter Ladeendspannung
#define E_AUX_AN		250000L		// mWh d.h. 1kWh=1000000
#define E_AUX_AUS		1000000L	// mWh d.h. 1kWh=1000000 
#define WERTE  0
#define GESAMT 1
#define AUS		0
#define AN		1

void drucke(int);					 
uint16_t messen();
char* pform(char*,char* label,uint8_t prec,int32_t,char* unit);
long formen(long wert,long unten,long oben);
long get_ibat_mA(void);
long get_ubat_mV(void);
long get_ures_mV(void);
long get_uvcc_mV(void);
long get_tcel_mK(void);
void action1(int intervall);
void action2(int intervall);
void parse(char* cmd);			
void varinit(void);

long uvcc_mV,ibat_mA,ch_uAh,ubat_mV,ures_mV,en_mWh;	// Strom,Ladung,Spannung,Energie
long tcel_mK; 
int Alarm=AUS,Aux=AUS;

int main (void){
	//wdt_disable();		// kommt zu spt - deshalb im Startcode 
	_delay_ms(500);			// bis Spannung hochgefahren
	boardinit();
	varinit();
	sei();
	send("\r\nBatMon luft");
	_delay_ms(500);			// bis Spannung hochgefahren
	while(1){					// ultraschnelle Schleife
		parse(cmd);				// sofortige Antwort
		action1(1);				// jede Sekunde zu erledigen
       	action2(3600);			// 1x pro Stunde
		}						
}							

void varinit(void){ 			// 100byte
	if (get_ubat_mV()>46000){
		en_mWh=(51000L-get_ubat_mV())*1000;	// 51-48V = 3000Wh (mit mWh-Prz.)
		ch_uAh=(en_mWh/50)*1000;			// 3000Wh@50V = 60Ah (mit uAh-Prz)
		}
	en_mWh=3170000L;			// Wert vorm Update restaurieren
	ch_uAh=67600000L;			// Wert vorm Update restaurieren
}

void parse(char * s){			
	if (strstr(s,"BM_Test")!=NULL){
		send("\r\nBM_Echo: ");
		send(s);
	//	_delay_ms(50);			// Sendestrung abklingen lassen
		s[0]='\0';
		}
	if(strstr(s,"BM_Reboot")!=NULL){
	    send("\n\rBM_Reboot - Springe zum Bootloader...");
		_delay_ms(10);
		wdt_enable(WDTO_30MS);while(1);	// WDT-Reset
		}
	if (strstr(s,"BM_Werte")!=NULL){
		drucke(WERTE);
	//	_delay_ms(50);			// Sendestrung abklingen lassen
		s[0]='\0';
		}
	if (strstr(s,"BM_ALTest")!=NULL){
		Alarm=AN;
		_AL_H;_AL_OUT;_delay_ms(100);_AL_IN;_AL_L; 	// Beeper AN
		send("\r\nBM_ALARM_AN");// der Raspi-Zentrale melden
	//	_delay_ms(50);			// Sendestrung abklingen lassen
		s[0]='\0';
		}
}

void action1(int intervall){	
static int i=0; static uint8_t ov=1; uint8_t nv; 
	if (((nv=tick)==0)&&(ov==1)) i++; ov=nv;// s-Takt
	if (i==intervall){	i=0;				// Intervall
		tcel_mK=get_tcel_mK();
		uvcc_mV=get_uvcc_mV();
		ures_mV=get_ures_mV();
		ubat_mV=get_ubat_mV(); 
		ibat_mA=get_ibat_mA(); 
		
		if (ures_mV<500){
	     	send("\n\rReset-Taste - Springe zum Bootloader...");
		 	_delay_ms(10);
		 	wdt_enable(WDTO_30MS);while(1);	// WDT-Reset			
			}
		if (ubat_mV<UBAT_FULL){	// Sekundentakt
			en_mWh+=((ibat_mA/10)*(ubat_mV/100))/3600; // mWh  max.4Mio
			ch_uAh+=(ibat_mA*1000)/3600; // Zehntel Ah  z.B. max 90Mio
			}
		else{
			en_mWh=0;				// Batterie voll - Reset
			ch_uAh=0;				// Batterie voll - Reset
			}
		if ((en_mWh<E_AUX_AN)&(Aux==AUS)) {_AUX_H; Aux=AN; send("\r\nBM_AUX_AN");send("\r\nTM_Switch21");}
		if ((en_mWh>E_AUX_AUS)&(Aux==AN)) {_AUX_L; Aux=AUS; send("\r\nBM_AUX_AUS");send("\r\nTM_Switch20");}
		if ((_ALState==1)&(Alarm==AUS)) {Alarm=AN; send("\r\nBM_ALARM_AN");}
		if ((_ALState==0)&(Alarm==AN)) {Alarm=AUS; send("\r\nBM_ALARM_AUS");}
		
		if ((ubat_mV<UBAT_MIN_AL)|(ubat_mV>UBAT_MAX_AL)| \
			(ibat_mA<IBAT_CHR_AL)|(ibat_mA>IBAT_DCH_AL)| \
			(tcel_mK>TEMP_AL)){
			if (Alarm==AUS) {
				Alarm=AN;
				_AL_H;_AL_OUT;_delay_ms(100);_AL_IN;_AL_L; 	// Beeper AN
				send("\r\nBM_ALARM_AN");					// der Zentrale melden
				}					
			}
		}
}

void action2(int intervall){				// luft auerhalb der ISR
static int i=0; static uint8_t ov=1; uint8_t nv; 
	if (((nv=tick)==0)&&(ov==1)) i++; ov=nv;// s-Takt
	if (i==intervall){ i=0;					// Intervall
		drucke(WERTE);						// Variablen drucken
		}
}

void drucke(int was){					// Leerzeichen als Abstandshalter
char zk[32];
	if (was==WERTE){	
		send("\r\n");
		send(pform(zk,"Ch: ",1,ch_uAh/1000," Ah "));	// max.90Ah 
		send(pform(zk,"En: ",2,en_mWh/1000," kWh "));// max.4kWh 
		send(pform(zk,"Pw: ",0,( (ibat_mA/32)*(ubat_mV/31) )  ," W ")); // mW
		send(pform(zk,"I: ",1,ibat_mA," A "));		// A
		send(pform(zk,"U: ",2,ubat_mV," V "));		// V
		send(pform(zk,"T: ",1,tcel_mK," C ")); 		// C 
		}
	if (was==GESAMT){
		//send("\n\rGesamtertrag: ");send(pform(zk,2,en_mWh/1000));send("kWh \n\r");  // max.20kWh pro Tag 20.000.000.00
		}
}

long get_ibat_mA(void){
long PB_mV,I_A;
	ADCSRA = 0;								// ADC Aus
	ADMUX = (1<<MUX1);						// Vccref,10bit,PB4 ADC2
	PB_mV=(uvcc_mV*messen()) >> 16; 		// ggf. falsch bei neg. Werten
	I_A=(((PB_mV-uvcc_mV/2)*1000)/ISENS_uV_pro_mA) + I_OS_mA ;
	return (-I_A*I_Adjust)/1000;			// evtl. noch Scale-Kalib
}

long get_ubat_mV(void){						
long PB_mV,Bat_mV;
	ADCSRA = 0;								// ADC Aus
	ADMUX = 0;								// Vccref,10bit,PB5 ADC0 
	PB_mV=(uvcc_mV*(long)messen()) >> 16;	// 10ms messen
	PB_mV -= ((uvcc_mV-PB_mV)*42)/520;		// Reset Pull-Up 52k korrigieren
	Bat_mV=(PB_mV*(42+517))/42;				// 47k+4k7 gegen 4k2
	return  (Bat_mV-UBAT_OS_mV);			// Offset-Korrektur
}

long get_ures_mV(void){					
	ADCSRA = 0;								// ADC Aus
	ADMUX = (1<<MUX0);						// Vccref,10bit,PB2 ADC1
	return ((uvcc_mV*(long)messen()) >> 16);// 10ms messen
}

long get_uvcc_mV(void){						
	ADCSRA = 0;								// ADC Aus
	ADMUX = (1<<MUX3)|(1<<MUX2);			// Vccref,10bit,Vbg
	return ((Vbg_mV<<16)/(long)messen());	// 10ms messen
}

long get_tcel_mK(void){								
long val=0;							
	ADCSRA = 0;								// ADC Aus
	ADMUX = (1<<REFS1)|(1<<MUX3)|(1<<MUX2)|(1<<MUX1)|(1<<MUX0);// 1.1Vref,int.T-Sensor ADC4
	val= ((Vbg_mV*10*(long)messen()) >> 16);// 10ms messen
	return (val-3000)*100;						// K-genau long lst aber mK auf	
}


uint16_t messen(void){						// 16bit Auflsung
#define Loops 100							// 10ms Messzeit kaum weisses Rauschen
uint32_t accu=0,i=0;						// bis 100 loops 16bit Auflsung

	ADCSRA= (1<<ADEN)|(1<<ADSC)|(1<<ADPS2)|(1<<ADPS1);//An,Start,Einmal-Modus,Takt/64
	loop_until_bit_is_clear(ADCSRA, ADSC); 	// Warten bis Fertig 
	while (i++<Loops){
	    ADCSRA |= (1<<ADSC); 					// Start ADC
    	loop_until_bit_is_clear(ADCSRA, ADSC); 	// Warten bis Fertig 13Takte =104s
		accu += ADCW;							// Ergebnis lesen
		}
	accu<<=6;								// 6bit schieben - tun so als ob wir 16bit htten
	return (uint16_t)(accu/=Loops);			// 16bit Ergebnis
}

//Formatierung Zielstring[min12], Nachkommastellen,32bit Signed-Long-Wert 
char* pform(char * zk, char* s1,uint8_t prec, int32_t v,char* s2) { 
uint8_t i,j=0,f,kpos=4;	// i:tempindex, j:outindex, flag und prec=4 fr mK-Genauigkeit
char s[32];
	
	if(v<0) { s[j++]='-'; v=-v;} 			 // Vorzeichen
	for(i=10;i>=1;i--){s[i]=(v%10)+'0';v/=10;}// Modulo rechnen, dann den ASCII-Code von '0' addieren
	for(i=1,f=0;i<(12-kpos);i++) {  				
		if (f==1 || s[i]!='0') {				// einfach Hochzhlen bei bei fhrenden Nullen
			s[j++]=s[i];						// relvante Vorkommastellen
			f=1;								// ... erreicht
			}
		}
	if (prec>0){
		if (f==0) s[j++]='0'; 					// eine fhrende Null
		s[j++]='.';                				// Komma ausgeben
		for(; i<(12-kpos+prec); i++) s[j++]=s[i];	// Nachkommastellen ausgeben
		}
	else{										// 0 Watt - Bug
		if (f==0) s[j++]='0'; 					// eine fhrende Null
		}
	s[j]='\0';                  				// String Terminator
	strcpy(zk,s1);
	strcat(zk,s);
	strcat(zk,s2);
	return zk;
}

uint8_t mcusr_mirror __attribute__ ((section (".noinit")));

void get_mcusr(void) __attribute__((naked)) __attribute__((section(".init1")));
void get_mcusr(void){
  mcusr_mirror = MCUSR;
  MCUSR = 0;		// disable the dog
  wdt_disable();
}
