#include <avr/io.h> 
#include <avr/wdt.h>
#include <util/delay.h>			// 8MHz ntig
#include <avr/interrupt.h>
#include <string.h>
#include "board.h"

uint8_t rxtest (uint8_t);	
uint8_t ugets(char*);
void uputs(char *);				// von parse verwendet
void uputc(char);
char ugetc(void);
char samplechar(void);

void boardinit (void){
	PORTB=0b00000000;      // Port initialisieren, 0=kein Pullup
    DDRB= 0b00000001;		// -,-,-,Isens,Usens,Alarm,Funk,Switch
	PCMSK |= (1<<_RXPin);	// PORTBx=PCINTx=_RX_Pin
	GIMSK |= (1<<PCIE);		// enable pin change interrupt

	TCCR1 |= 0x8F;			// CTC1, Systemtakt/16384
	OCR1A = 0;				// interrupt z.B. bei 0 
	OCR1C = 246;			// 244 0.5Hz Takt
	TIMSK |= (1<<OCIE1A);	// enable compareA-Interrupt
}
// wdt_init - Watchdog Init used to disable the CPU watchdog placed in Startcode, no call needed
#include <avr/wdt.h>
void wdt_init(void) __attribute__((naked)) __attribute__((section(".init1")));
void wdt_init(void){
    MCUSR = 0;
    wdt_disable();
    return;
}

ISR (TIM1_COMPA_vect){		
	tick ^= 0x01;			// 1Hz-Takt
}

ISR (PCINT0_vect){			// sendet oder lauscht auf Befehl, typ 10ms max 60ms Laufzeit 
	PCMSK &= ~(1<<_RXPin);	// PORTBx=PCINTx  pcintx on PBx deaktivieren
	ugets(cmd);				// Anzahl der Bytes 
	PCMSK |= (1<<_RXPin);	// PORTBx=PCINTx  pcintx on PBx aktivieren
}

uint8_t ugets(char * s){	// kompletten String lesen, CR oder Timeout 
uint8_t i=0,c;
	while (i<STRING_SIZE){
	 	c=ugetc();
	 	if (c=='\r') 
			c='\0';
		s[i++]=c;
		if (c=='\0')
			return i;		// Anzahl der gelesenen Zeichen vor Timeout oder CR
		}
	s[0]='\0';
	return 0;				// zu langer String ist immer Schrott
}			

void send(char* cp){				// ich lasse! von der IRS senden
	PCMSK &= ~(1<<_RXPin);	// PORTBx=PCINTx  pcintx on PBx deaktivieren
//	_delay_ms(10);
	uputs(cp);	
	_delay_ms(50);			// Sendestrung abklingen lassen
	PCMSK |= (1<<_RXPin);	// PORTBx=PCINTx  pcintx on PBx aktivieren
}	

char ugetc(void){			// warte bis Startbit und lese Zeichen
     uint32_t i=0;
	 char c;
     while ((rxtest(_RXState)==0)&&(i<_TOUT)){ i++; } // warte auf nchstes Zeichen
     if (i<_TOUT)
	    c=samplechar();		// Zeichen lesen
	 else  
		c='\0';
	 return c;
}

char samplechar(void){		// lese Zeichen
uint8_t c=0,i;

	_delay_us(_TBIT/4);     // Mitte vom Start-Bit _TBIT/2 minus Laufzeiten
    for (i=0;i<8;i++){		// bit0 bis bit7 
        _delay_us(_TBIT);
        if (_RXState==1)	// abtasten
		   c|=(1<<i);       // und ggf. Bit setzen   
	    }
    _delay_us(_TBIT);       // Stop-Bit abwarten
	return c;
}

void uputs(char * s){				// sendet String s
	while (*s!='\0') uputc(*s++);
}

void uputc(char c){      // 1Start 8Data 1Stop
    uint8_t i,buff[10];
    
	_TXPin_H;
	_TXMode;
    buff[0]=0;				// Start-Bit
    for (i=1;i<9;i++){
		buff[i]=c&0b00000001;
		c>>=1;
        }
    buff[9]=1; 				// Stopp-Bit          
    for (i=0;i<10;i++){
        if (buff[i]==0)
			_TXPin_L;
        else 
		   	_TXPin_H;
		_delay_us(_TBIT);
	    }
    _TXPin_H;
	_RXMode;
}

uint8_t rxtest (uint8_t nv){	
static uint8_t ov=1;
uint8_t	rv=0;
	if ((nv==0)&&(ov==1))
        rv=1;
	ov=nv;
    return rv;
}


